#!/usr/bin/env python3

"""

 Functions that have no home...

"""

import re
import os
import csv
import yaml
import glob
import json
import shutil
import pickle
import subprocess as sub
from pathlib import Path

# Module imports
from pyinseq.logger import pyinseq_logger

logger = pyinseq_logger.logger


def create_experiment_directories(settings):
    """
    Create the project directory and subdirectories

    Attempt to create the directory structure:

    results/
    |
    +-{settings.experiment}/        # User-specified experiment name
      |
      +-raw_data/          # For demultiplexed reads
      |
      +-genome_lookup/     # Genome fna and ftt files, bowtie indexes

    If /experiment directory already exists exit and return error message and
    the full path of the present directory to the user"""
    # Check that experiment name has no special characters or spaces
    experiment = convert_to_filename(settings.experiment)
    output_path = settings.path

    # ERROR MESSAGES
    error_directory_exists = (
        f"Directory {experiment} already exists. "
        f"Snakemake will create files that are missing."
    )

    try:
        output_path.mkdir(parents=True, exist_ok=False)
        logger.info(f"Make directory: results/{experiment}")

        if settings.process_reads:
            # Raw data dir
            output_path.joinpath("raw_data").mkdir()
            logger.info(f"Make directory: results/{experiment}/raw_data/")
        if settings.parse_genebank:
            # Genome dir
            output_path.joinpath("genome_lookup").mkdir()
            logger.info(f"Make directory: results/{experiment}/genome_lookup/")
    except FileExistsError:
        logger.info(error_directory_exists)
        return


def convert_to_filename(sample_name):
    """
    Convert to a valid filename.

    Removes leading/trailing whitespace, converts internal spaces to underscores.
    Allows only alphanumeric, dashes, underscores, unicode.
    """
    return re.sub(r"(?u)[^-\w]", "", sample_name.strip().replace(" ", "_"))


def count_lines(filename: str, fastq=False):
    """
    Returns number of lines in given file for progress bar.
    Uses subprocess to call wc in shell
    """
    try:
        # Check if gz compressed
        if filename.split(".")[-1] in ["gz"]:
            # Build pipe for uncompressed file
            pipe = sub.Popen(("gzcat", "-cd", filename), stdout=sub.PIPE)
            output = int(sub.check_output(["wc", "-l"], stdin=pipe.stdout))
        # Other files...
        else:
            output = int(
                sub.check_output(["wc", "-l", filename], stderr=sub.DEVNULL)
                .split()[0]
                .decode("UTF-8")
            )
        if fastq:
            return int(output / 4)
        else:
            return int(output)
    except sub.CalledProcessError as e:
        # Catching error returns arbitrary value
        logger.debug(f"{str(e)}: using arbitrary limit for progress bar")
        return None


def read_gene_file(gene_file):
    """
    Helper for reading a sample gene file.
    Returns a dictionary where key is gene and value is cpm

    :param gene_file:
    :return:
    """
    gene_dict = dict()
    with open(gene_file, "r") as f:
        reader = csv.DictReader(f, delimiter="\t")
        for record in reader:
            gene_dict[record["locus_tag"]] = [float(record["cpm"])]

    return gene_dict


def get_version():
    """
    Returns pyinseq version
    :return:
    """
    dir = os.path.dirname(__file__)
    with open(os.path.join(dir, "VERSION"), "r") as f:
        return f.read()


def get_config_dict(config_file):
    """Loads data from config_file into dict"""
    with open(config_file, "r") as f:
        return yaml.load(f, Loader=yaml.FullLoader)


def write_config_file(args=None, format="yml", default=False) -> Path:
    """

    Given pyinseq arguments, creates a config file that can be given to snakemake

    :param args: Namespace object with pyinseq args
    :param format: file extension for config_file (json or yaml)
    :param default: boolean to write default config file
    :return: path to config_file
    """
    args_dict = vars(args)
    dumper = {"yml": yaml.dump, "json": json.dump}[format]
    if default:
        logger.info("Writing DEFAULT configuration file as template for pyinseq")
        with open(f"default-config-pyinseq.{format}", "w") as f:
            config_dict = {key: args_dict[key] for key in args_dict}
            dumper(config_dict, f)
        logger.info("Exiting gracefully...")
        exit(0)
    else:
        logger.info("Writing configuration file from provided arguments")
        config_file = Path(f"{args.experiment}-config.{format}")
        with open(config_file, "w") as f:
            if not args_dict["command"]:
                args_dict["command"] = "pyinseq"
            config_dict = {key: args_dict[key] for key in args_dict}
            dumper(config_dict, f)
        return config_file


def tab_delimited_samples_to_dict(sample_file):
    """Read sample names, barcodes from tab-delimited into an OrderedDict."""
    samples_dict = {}
    with open(sample_file, "r", newline="") as csv_file:
        for line in csv.reader(csv_file, delimiter="\t"):
            # ignore comment lines in original file
            if not line[0].startswith("#"):
                # sample > filename-acceptable string
                # barcode > uppercase
                sample = convert_to_filename(line[0])
                barcode = line[1].upper()
                if sample not in samples_dict and barcode not in samples_dict.values():
                    samples_dict[sample] = {"barcode": barcode}
                else:
                    raise IOError(f"Error: duplicate sample {sample} barcode {barcode}")
    return samples_dict


def yaml_samples_to_dict(sample_file):
    """Read sample names, barcodes from yaml."""
    with open(sample_file, "r") as f:
        samples_dict = yaml.load(sample_file)
    return samples_dict


def directory_of_samples_to_dict(directory):
    """Read sample names from a directory of .gz files."""
    samples_dict = {}
    for gz_file in list_files(directory):
        # TODO(convert internal periods to underscore? use regex?)
        # extract file name before any periods
        f = os.path.splitext(os.path.basename(gz_file))[0].split(".")[0]
        samples_dict[f] = {}
    return samples_dict


def list_files(folder, ext="gz"):
    """Return list of .gz files from the specified folder."""
    with cd(folder):
        return [f for f in glob.glob(f"*.{ext}")]


"""

Functions that handle terminal operations

"""


def copy_file(src, dest):
    """Literally does what it says..."""
    shutil.copy(src, dest)


def execute(cmd) -> None:
    """Execute snakemake workflow using subprocess module"""
    # Call bash command, suppress pyinseq output
    if not isinstance(cmd, list):
        cmd = cmd.split(" ")
    process = sub.Popen(cmd)
    try:
        process.wait()
    except KeyboardInterrupt:
        # Kill if CTRL+C event occurs
        process.kill()
    finally:
        process.terminate()


def pickle_object(object, pickle_output):
    """Serializes objects that can be picked up by other processes"""
    with open(pickle_output, "wb") as f:
        pickle.dump(object, f)


def load_pickle(byte_file):
    """Loads a pickled object"""
    with open(byte_file, "rb") as f:
        return pickle.load(f)



#!/usr/bin/env python3

"""Configuration file for pyinseq.py package."""
import os
import sys

# TRANSPOSON END SEQUENCES
# Typical mariner transposon ends for INSeq are the defaults (same left/right ends)
# This sequence is filtered from the reads prior to mapping
# Mariner transposons inserts at TA dinucleotides; do not include the corresponding
# TA from the tranpsoson here
transposon_left = "ACAGGTTG"
transposon_right = "ACAGGTTG"

# Bowtie package selected (linux/mac/windows) based on operating system detected
platform = sys.platform.lower()
current_folder = os.path.dirname(os.path.abspath(__file__))
packages_folder = os.path.join(current_folder, "third-party")

default_version = "bowtie-1.1.1-mac"

if platform.startswith("linux"):
    # linux bowtie path
    default_version = "bowtie-1.1.1-linux"

elif platform.startswith("win"):
    # windows bowtie path
    default_version = "bowtie-1.1.1-win"

bowtie = os.path.join(packages_folder, default_version, "bowtie")

# PATH TO BOWTIE-BUILD (appends '-build' on the path above)
bowtieBuild = f"{bowtie}-build"



#!/usr/bin/env python3
import os
import sys
import filecmp
import traceback
from distutils import dir_util
from pathlib import Path

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

import pytest


class cd:
    """Context manager to change to the specified directory then back."""

    def __init__(self, new_path):
        self.new_path = os.path.expanduser(new_path)

    def __enter__(self):
        self.savedPath = os.getcwd()
        os.chdir(self.new_path)

    def __exit__(self, etype, value, traceback):
        os.chdir(self.savedPath)


def run_pytest():
    """Runs pytest directly from source code. CWD should be in pyinseq source dir"""
    rootdir = Path(__file__).parents[2]
    with cd(rootdir):
        pytest.main([])


def get_dump():
    """Helper for creating and returning a dump Path"""
    dump = Path("pyinseq/tests/dump")
    if not dump.exists():
        dump.mkdir()
    return dump


@pytest.fixture()
def load_settings():
    def fetch(command):
        config_file = "../data/input/test-pyinseq-config.yml"
        dump = get_dump()
        with cd(dump):
            from pyinseq.settings import SETTINGS_CONSTRUCTORS

            settings_constructor = SETTINGS_CONSTRUCTORS[command]
            return settings_constructor(config_file)

    return fetch


"""
These script running functions were taken from the khmer project:
https://github.com/dib-lab/khmer/blob/master/tests/khmer_tst_utils.py
"""


@pytest.fixture(scope="session")
def datadir(tmpdir_factory, request):
    """
    Fixture responsible for locating the test data directory and copying it
    into a temporary directory.
    """
    tmpdir = tmpdir_factory.mktemp("data")
    # this is the data/ dir in pyinseq/tests
    data_dir = os.path.join(os.path.dirname(__file__), "data")
    dir_util.copy_tree(data_dir, str(tmpdir))

    # Create dump in temporary directory
    tmpdir.join("dump").mkdir()

    def getter(filename="", as_str=True):
        filepath = tmpdir.join(filename)
        if as_str:
            return str(filepath)
        return filepath

    return getter


def scriptpath(scriptname="pyinseq"):
    "Return the path to the scripts, in both dev and install situations."

    path = os.path.join(os.path.dirname(__file__), "../../scripts")
    if os.path.exists(os.path.join(path, scriptname)):
        return path

    path = os.path.join(os.path.dirname(__file__), "../../../EGG-INFO/scripts")
    if os.path.exists(os.path.join(path, scriptname)):
        return path

    for path in os.environ["PATH"].split(":"):
        if os.path.exists(os.path.join(path, scriptname)):
            return path


def _runscript(scriptname):
    """
    Find & run a script with exec (i.e. not via os.system or subprocess).
    """

    import pkg_resources

    ns = {"__name__": "__main__"}
    ns["sys"] = globals()["sys"]

    try:
        pkg_resources.get_distribution("pyinseq").run_script(scriptname, ns)
        return 0
    except pkg_resources.ResolutionError as err:
        path = scriptpath()

        scriptfile = os.path.join(path, scriptname)
        if os.path.isfile(scriptfile):
            if os.path.isfile(scriptfile):
                exec(compile(open(scriptfile).read(), scriptfile, "exec"), ns)
                return 0

    return -1


def runscript(
    scriptname: str, args: list, directory=None, fail_ok=False, sandbox=False
) -> [str, str, str]:
    """Run a Python script using exec().
    Run the given Python script, with the given args, in the given directory,
    using 'exec'.  Mimic proper shell functionality with argv, and capture
    stdout and stderr.
    When using :attr:`fail_ok`=False in tests, specify the expected error.
    """
    sysargs = [scriptname]
    sysargs.extend(args)
    cwd = os.getcwd()
    print(os.getcwd())

    try:
        status = -1
        oldargs = sys.argv
        sys.argv = sysargs

        oldout, olderr = sys.stdout, sys.stderr
        sys.stdout = StringIO()
        sys.stdout.name = "StringIO"
        sys.stderr = StringIO()

        # Use with to change directories
        if not directory:
            directory = cwd

        with cd(directory):
            try:
                print("running:", scriptname, "in:", directory, file=oldout)
                print("arguments", sysargs, file=oldout)
                status = _runscript(scriptname)
            except SystemExit as e:
                status = e.code
            except:
                traceback.print_exc(file=sys.stderr)
                status = -1
    finally:
        sys.argv = oldargs
        out, err = sys.stdout.getvalue(), sys.stderr.getvalue()
        sys.stdout, sys.stderr = oldout, olderr
        print(out)

    if status != 0 and not fail_ok:
        print(
            "Script Failed:",
            scriptname,
            "Status:",
            status,
            "Output:",
            out,
            "Error:",
            err,
            sep="\n",
        )
        assert False
    return status, out, err


def compare_directories(expected_output, output_dir, ignore=[]):
    """Compare output from pyinseq runs to expected output"""
    dcmp = filecmp.dircmp(expected_output, output_dir, ignore=ignore)
    # check that log file is created from pyinseq
    assert "log.txt" in os.listdir(output_dir)

    # checks that files are same in both directories
    assert not dcmp.left_only and not dcmp.right_only

    # check files to see if content differs
    assert not dcmp.diff_files
    assert not dcmp.funny_files  # Check for files that cannot be compared

    # because subdirs is a dict keyed by subdir name
    # with dircmp objects as values
    for subdcmp in dcmp.subdirs.values():
        assert not subdcmp.diff_files
        assert not subdcmp.funny_files
        assert not subdcmp.left_only and not subdcmp.right_only



#!/usr/bin/env python3

"""

Contains functions for parsing arguments from command line

"""

import os
import argparse

# Module imports
from pyinseq.utils import get_version


def get_snake_parser():
    # Snake parser will be parent for all other parsers
    snake_parser = argparse.ArgumentParser(add_help=False)
    snake_group = snake_parser.add_argument_group("SNAKEMAKE")
    snake_group.add_argument(
        "--get_default_config",
        action="store_true",
        help="Writes a default configuration file for pyinseq run that can be modified",
    )
    snake_group.add_argument(
        "--config_format",
        default="yml",
        help="Format for the configuration file. Options are: 'json' or 'yml'",
    )
    snake_group.add_argument(
        "-c",
        "--config",
        help="Provide config file (in json or yaml format) instead of pyinseq arguments to run pipeline",
        required=False,
    )
    snake_group.add_argument(
        "-t",
        "--threads",
        default=os.cpu_count(),
        help="Number of threads that snakemake can use to run pyinseq, the more the better for parallel processing",
        type=int,
    )
    snake_group.add_argument(
        "--snakemake_params",
        help="Additional params passed to snakemake. "
        "This should be included at the end of the command since everything will be passed to snakemake so make sure they are correct. "
        "For example, you can use `-n` to check which files snakemake will create without execution the full workflow.",
        nargs="...",
        default=[],
        type=str,
    )
    snake_group.add_argument(
        "--test",
        action="store_true",
        help="Runs pytest on pyinseq module",
        required=False,
        default=False,
    )
    return snake_parser


def get_args():
    """Parse command line arguments for main pyinseq. Returns both parser and Namespace object"""

    parser = argparse.ArgumentParser("pyinseq", parents=[get_snake_parser()])
    parser.add_argument(
        "-v", "--version", action="version", version=f"pyinseq: {get_version()}"
    )
    parser.add_argument(
        "-i", "--input", help="input Illumina reads file or folder", required=False
    )
    parser.add_argument(
        "-s", "--samples", help="sample list with barcodes", required=False
    )
    parser.add_argument(
        "-e",
        "--experiment",
        help="experiment name (no spaces or special characters)",
        required=False,
    )
    parser.add_argument(
        "-g",
        "--genome",
        help="genome in GenBank format (one concatenated file for multiple contigs/chromosomes)",
        required=False,
    )
    parser.add_argument(
        "-d",
        "--disruption",
        help="fraction of gene disrupted (0.0 - 1.0). Default: 0.9",
        default=0.9,
        type=float,
    )
    parser.add_argument(
        "--min_count",
        help="Minimum number of reads per insertion site",
        default=3,
        type=int,
    )
    parser.add_argument(
        "--max_ratio",
        help="Maximum ratio of left:right or right:left reads per insertion site",
        default=10,
        type=int,
    )
    parser.add_argument(
        "--barcode_length",
        help="Length of the barcode which is used to demultiplex samples (4 - 16bp)",
        type=int,
        default=4,
    )
    parser.add_argument(
        "--transposon_seq",
        help="Sequence for the transposon that flanks reads",
        default="ACAGGTTG",
    )
    parser.add_argument(
        "--gff3",
        help="generate GFF3 file",
        action="store_true",
        required=False,
        default=False,
    )
    subparsers = parser.add_subparsers(
        dest="command", help="Command for performing specialized tasks"
    )
    # demultiplex
    sub_parser_demultiplex = subparsers.add_parser(
        "demultiplex",
        parents=[get_snake_parser()],
        help="Demultiplex reads into barcode samples",
    )
    sub_parser_demultiplex.add_argument(
        "-i", "--input", help="input Illumina reads file or folder", required=False
    )
    sub_parser_demultiplex.add_argument(
        "-s", "--samples", help="sample list with barcodes", required=False
    )
    sub_parser_demultiplex.add_argument(
        "-e",
        "--experiment",
        help="experiment name (no spaces or special characters)",
        required=False,
    )
    sub_parser_demultiplex.add_argument(
        "--notrim",
        help="do not write trimmed reads (i.e. write raw reads only)",
        action="store_true",
        required=False,
        default=False,
    )
    sub_parser_demultiplex.add_argument(
        "--barcode_length",
        help="Length of the barcode which is used to demultiplex samples (4 - 16bp)",
        type=int,
        default=4,
    )
    sub_parser_demultiplex.add_argument(
        "--transposon_seq",
        help="Sequence for the transposon that flanks reads",
        default="ACAGGTTG",
    )

    # Genomeprep
    sub_parser_genome_prep = subparsers.add_parser(
        "genomeprep",
        parents=[get_snake_parser()],
        help="Prepare genome files from nucleotide sequences",
    )
    sub_parser_genome_prep.add_argument(
        "-e",
        "--experiment",
        help="experiment name (no spaces or special characters)",
        required=False,
    )
    sub_parser_genome_prep.add_argument(
        "-g",
        "--genome",
        help="genome in GenBank format (one concatenated file for multiple contigs/chromosomes)",
        required=False,
    )
    sub_parser_genome_prep.add_argument(
        "--noindex",
        help="do not generate bowtie indexes",
        action="store_true",
        required=False,
        default=False,
    )
    sub_parser_genome_prep.add_argument(
        "--gff3",
        help="generate GFF3 file",
        action="store_true",
        required=False,
        default=False,
    )
    return parser, parser.parse_args()


"""Inactive arguments in current version
parser.add_argument('--nobarcodes',
                    help='barcodes have already been removed from the samples; \
                    -i should list the directory with filenames (.fastq.gz) \
                    corresponding to the sample names',
                    action='store_true',
                    default=False)
parser.add_argument('--compress',
                    help='compress (gzip) demultiplexed samples',
                    action='store_true',
                    default=False)
parser.add_argument('--keep_all',
                    help='keep all intermediate files generated',
                    action='store_true',
                    default=False)"""



#!/usr/bin/env python3

"""

Main script for running the pyinseq package

"""

import yaml

# Module imports
from pyinseq.utils import copy_file
from pyinseq.analyze import t_fifty
from pyinseq.settings import Settings
from pyinseq.logger import pyinseq_logger
from pyinseq.map_reads import bowtie_map, parse_bowtie
from pyinseq.process_mapping import map_sites, map_genes, build_gene_table

logger = pyinseq_logger.logger


def pipeline_mapping(settings, samples_dict):
    """Aggregate bowtie output, map to genes in the feature table, and aggregate samples."""
    # Dictionary of each sample's cpm by gene
    gene_mappings = {}
    mapping_data = {}
    for sample in samples_dict:
        with cd(settings.genome_path):
            # Paths are relative to the genome_lookup directory
            # from where bowtie is called
            bowtie_in = "../" + sample + "_trimmed.fastq"
            bowtie_out = "../" + sample + "_bowtie.txt"
            # map to bowtie and produce the output file
            logger.info(f"Sample {sample}: map reads with bowtie")
            bowtie_msg_out = bowtie_map(settings.organism, bowtie_in, bowtie_out)
            logger.info(bowtie_msg_out)
            # store bowtie data for each sample in dictionary
            mapping_data[sample] = {"bowtie_results": [], "insertion_sites": []}
            mapping_data[sample]["bowtie_results"] = parse_bowtie(bowtie_msg_out)
        # Map each bowtie result to the chromosome
        insertions = len(map_sites(sample, settings))
        mapping_data[sample]["insertion_sites"] = insertions
        # Add gene-level results for the sample to geneMappings
        # Filtered on gene fraction disrupted as specified by -d flag
        gene_mappings[sample] = map_genes(sample, settings)
        # if not settings.keep_all:
        #    # Delete trimmed fastq file, bowtie mapping file after writing mapping results
        #    os.remove(s["trimmedPath"])
        #    os.remove("results/{Settings.experiment}/{bowtieOutputFile}"
        t_fifty_result = t_fifty(sample, settings)
        logger.info(f"T50 result for {sample}: {t_fifty_result}")
    build_gene_table(
        settings.organism, samples_dict, gene_mappings, settings.experiment
    )


def pipeline_summarize(samples_dict: dict, settings: Settings):
    """Summary of INSeq run."""
    # Write summary log with more data
    copy_file(settings.samples, settings.samples_txt)
    logger.info(f"Print log path: {settings.log}")
    pyinseq_logger.logger_io.write(f"Print log path: {settings.log}\n")
    logger.info(f"Print settings:\n{settings}\n")
    pyinseq_logger.logger_io.write(f"Print settings:\n{settings}\n")
    with open(settings.samples_info_yml, "r") as f:
        d = yaml.load(f, Loader=yaml.FullLoader)
    logger.info(f"Print samples detail\n{yaml.dump(d, default_flow_style=False)}\n")


def pipeline_analysis(samples_dict: dict, settings: Settings) -> None:
    """Pipeline for analyzing pyinseq results"""
    # T50 calculation
    T50_dict = dict()
    for sample in samples_dict:
        T50_dict[sample] = t_fifty(sample, settings)
        logger.info(f"T50 {sample}: {T50_dict[sample]}")
        # plot_insertions(sample, settings)
    return T50_dict


if __name__ == "__main__":
    pass



#!/usr/bin/env python3

"""

Count and process the bowtie hits at each position in each sample

"""

import csv

# Module Imports
from pyinseq.logger import pyinseq_logger

logger = pyinseq_logger.logger


def map_sites(sample, settings):
    """Map insertions to nucleotide sites."""
    # Placeholder for dictionary of mapped reads in format:
    # {(contig, position) : [left_counts, right_counts]}
    map_dict = {}
    # overall_total = denominator for cpm calculation
    overall_total = 0
    bowtie_file = settings.path.joinpath(f"{sample}_bowtie.txt")
    sites_file = settings.path.joinpath(f"{sample}_sites.txt")
    logger.info(
        f"Sample {sample}: Tally site data from {bowtie_file.name} into {sites_file.name} file."
    )
    with open(bowtie_file, "r") as fi:
        for line in fi:
            bowtie_data = line.rstrip().split("\t")
            # Calculate transposon insertion point = insertion_NT
            contig, insertion_NT, read_length = (
                str(bowtie_data[2]),
                int(bowtie_data[3]),
                len(bowtie_data[4]),
            )
            if bowtie_data[1] == "+":  # positive strand read
                insertion_NT = insertion_NT + read_length - 1
                map_dict.setdefault((contig, insertion_NT), [0, 0])[0] += 1  # left
            else:  # negative strand read
                insertion_NT = insertion_NT + 1
                map_dict.setdefault((contig, insertion_NT), [0, 0])[1] += 1  # right
            overall_total += 1
    # write tab-delimited of contig/nucleotide/left_counts/right_counts/total_counts/cpm
    # use the total_counts as the denominator for cpm calculation
    with open(sites_file, "a") as fo:
        writer = csv.writer(fo, delimiter="\t", dialect="excel", lineterminator="\n")
        header_entry = (
            "contig",
            "nucleotide",
            "left_counts",
            "right_counts",
            "total_counts",
            "cpm",
        )
        writer.writerow(header_entry)
        sample_dict = {sample: {"site hits": 0}}
        for insertion in sorted(map_dict):
            left_counts = map_dict[insertion][0]
            right_counts = map_dict[insertion][1]
            total_counts = map_dict[insertion][0] + map_dict[insertion][1]
            cpm = float(1e6) * total_counts / overall_total
            row_entry = (
                insertion[0],
                insertion[1],
                left_counts,
                right_counts,
                total_counts,
                cpm,
            )
            writer.writerow(row_entry)
            sample_dict[sample]["site hits"] += total_counts
    # Summarize site data into io
    pyinseq_logger.logger_io.write(
        f"- {sample}: {overall_total} aligned reads mapped to sites\n"
    )
    # Return a dict where {sample: {sites: 100} }
    return sample_dict


def map_genes(sample, settings):
    """Maps insertions to genes

    1. Writes a csv file listing the gene for each insertion with the tabs:
    contig, nucleotide, Lcounts, Rcounts, totalCounts, cpm, threePrimeness, locus_tag
    2. Returns a dictionary of aggregate counts per gene (for any gene with at
    least one hit)

    Insertions that map to multiple genes result in multiple lines in the CSV file
    and are counted for both genes in the returned dictionary.

    ThreePrimeness = insertion location in gene (5' end = 0.0, 3' end = 1.0)
    All insertions are written to the file but only ones <= disruption threshold
    are counted in the dictionary.
    """
    # List of tuples of genome features
    genome = ftt_lookup(settings.organism, settings.experiment)
    # list of tuples of each mapped insertion to be immediately written per insertion
    mapped_hit_list = []
    # Dictionary with running total of cpm per gene; keys are genes, values are aggregate cpm
    # only hits in the first part of the gene are added to the count, as defined
    # by the disruption threshold.
    # if disruption = 1.0 then every hit in the gene is included
    gene_dict = {}
    sites_file = settings.path.joinpath(f"{sample}_sites.txt")
    genes_file = settings.path.joinpath(f"{sample}_genes.txt")
    second_message = f"Transposon hits that disrupt the 5-prime-most {round(settings.disruption * 100)}% of each gene are tallied"
    if settings.disruption == 1:
        second_message = (
            "Transposon hits that disrupt any position in a gene are tallied"
        )
    logger.info(
        f"Sample {sample}: Tally site data from {sites_file.name} to gene-level data in {genes_file.name}. {second_message}"
    )
    with open(sites_file, "r", newline="") as csvfileR:
        sites_reader = csv.reader(csvfileR, delimiter="\t")
        next(sites_reader, None)  # skip the headers
        sites_in_genes = 0
        for line in sites_reader:
            contig, nucleotide, left_counts, right_counts, total_counts, cpm = line[0:6]
            nucleotide = int(nucleotide)
            cpm = float(cpm)
            # Used to save previous feature for intergenic calling
            previous_feature = ""
            for gene in genome:
                (
                    locus,
                    start,
                    end,
                    strand,
                    length,
                    pid,
                    gene,
                    locus_tag,
                    code,
                    cog,
                    product,
                ) = (
                    gene[0],
                    int(gene[1]),
                    int(gene[2]),
                    gene[3],
                    int(gene[4]),
                    *gene[5:11],
                )
                # contig from insertion; locus from lookup table
                if contig == locus:
                    if start <= nucleotide <= end:
                        sites_in_genes += int(total_counts)
                        # 0.0 = 5'end ; 1.0 = 3'end
                        if strand == "+":
                            three_primeness = (1 + nucleotide - start) / length
                        else:
                            three_primeness = (end - nucleotide) / length
                        mappedHit = (
                            contig,
                            nucleotide,
                            left_counts,
                            right_counts,
                            total_counts,
                            cpm,
                            three_primeness,
                            locus_tag,
                        )
                        # Checks minimum count and max ratio
                        if int(total_counts) >= settings.min_counts and (
                            min(int(left_counts), int(right_counts))
                            * settings.max_ratio
                        ) >= max(int(left_counts), int(right_counts)):
                            mapped_hit_list.append(mappedHit)
                            # Filter based on location in the gene
                            if three_primeness <= settings.disruption:
                                # Add to the total for that gene --
                                # Single-element list (rather than integer) so
                                # that it is subscriptable to add cpm counts
                                gene_dict.setdefault(locus_tag, [0])[0] += cpm
                previous_feature = locus_tag
    # Write individual insertions to *_genes.txt
    sample_dict = {sample: {"gene hits": 0}}
    with open(genes_file, "w", newline="") as csvfileW:
        headers = (
            "contig",
            "nucleotide",
            "left_counts",
            "right_counts",
            "total_counts",
            "cpm",
            "three_primeness",
            "locus_tag",
        )
        mapped_gene_writer = csv.writer(csvfileW, delimiter="\t", lineterminator="\n")
        mapped_gene_writer.writerow(headers)
        for hit in mapped_hit_list:
            mapped_gene_writer.writerow(hit)
            sample_dict[sample]["gene hits"] += int(hit[4])
    # Summarize site data into io
    pyinseq_logger.logger_io.write(
        f"- {sample}: {sites_in_genes} of mapped reads fall in genes\n"
    )
    # Return aggregated insertions by gene (filtered on 5'-3' threshold)
    return sample_dict


def build_gene_table(organism, sample_dict, gene_mappings, experiment=""):
    """
    For each entry in a feature table (.ftt) list the summary of hits
    for each sample in the experiment
    """
    logger.info("Aggregate gene mapping from all samples into the summary_data_table")

    gene_table = ftt_lookup(organism, experiment)

    # Header will be extended in the future to
    # list the experiment and barcode of each sample of interest
    header = [
        "Contig",
        "Start",
        "End",
        "Strand",
        "Length",
        "PID",
        "Gene",
        "Synonym",
        "Code",
        "COG",
        "Product",
    ]

    # Add header row to ftt file
    gene_table.insert(0, header)

    # Get individual mapped hits
    # mappedHitList = mapToGene(organism, experiment)

    # current column, sample being matched
    current_column = len(gene_table[0]) - 1

    for sample in sample_dict:
        # Add the new sample name as a new column the table
        current_column += 1
        gene_table[0].append(sample)
        # Fill the rest of the new column with 0 as the count for each gene
        for row in gene_table[1:]:
            row.append(0)
        # add the sample's results to the building gene_table
        mapped_genes = gene_mappings[sample]
        # for each row in the table, *try* from mapped_genes. Add if found.
        for gene in mapped_genes:
            for i, f in enumerate(gene_table):
                hit_locus_tag = gene
                ftt_locus_tag = f[7]
                # matches based on locus_tag.
                if hit_locus_tag == ftt_locus_tag:
                    gene_table[i][current_column] += mapped_genes[gene][0]
        with open(f"results/{experiment}/summary_gene_table.txt", "w") as fo:
            writer = csv.writer(
                fo, delimiter="\t", dialect="excel", lineterminator="\n"
            )
            writer.writerows(gene_table)

    # Summarize build gene table step
    pyinseq_logger.logger_io.write(
        f"- Gene table contains {len(sample_dict)} samples (columns) "
        f"for {len(gene_table) - 1} genes (rows) \n"
    )
    return


def ftt_lookup(organism, experiment=""):
    """Import the ftt file and process as a dictionary of lookup values
    indexed on Synonym (i.e., Locus Tag)
    {'VF_0001': {'locus': 'CP000020', 'start': ...},
        'VF_0002': {'locus': 'CP000020', 'start': ...}}
    """
    ftt_list = []
    with open(
        f"results/{experiment}/genome_lookup/{organism}.ftt", newline=""
    ) as csv_file:
        ftt_reader = csv.reader(csv_file, delimiter="\t")
        for line in ftt_reader:
            # ignore header row
            if line[0] != ("Locus"):
                # Locus, Location_Start, Location_End, Strand, Length, PID,
                # Gene, Synonym, Code, COG, Product
                feature_data = line[0:11]
                ftt_list.append(feature_data)
    return ftt_list


def main():
    """Start here."""
    pass


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

"""
Demultiplexes a FASTQ file into multiple files by 5' end barcode.

Output path includes the experiment and sample name:
(pyinseq)/results/{experiment}/{sample}.fastq

"""

import re
import screed

# Module imports
from pyinseq.utils import count_lines
from pyinseq.logger import pyinseq_logger, TqdmBarLogger

logger = pyinseq_logger.logger


def demultiplex_fastq(reads, samples_dict: dict, settings) -> int:
    """Demultiplex a fastq input file by 5' barcode into separate files.

    Use regex to identify the chromosome slice and save this in the read
    record as 'trim', e.g., read['trim'] = (4, 21)
    Save raw reads into '{experiment}/raw_data/{sample_name}.fastq'
    Save trimmed reads into '{experiment}/{sample_name}_trimmed.fastq'

    :reads: path to file with reads
    :samples_dict: dict where sample is key and value={"barcode": "NNNN"}
    :settings: Settings object for experiment
    :return: a dict where key=barcode and value=num of reads
    """
    # Dictionary of lists to hold FASTQ reads until they are written to files
    # Keys are barcodes
    demultiplex_dict = {}
    # Keys are barcodes but used to count reads in sample
    demultiplex_count_dict = {}
    for sample in samples_dict:
        bc = samples_dict[sample]["barcode"]
        demultiplex_dict[bc] = []
        demultiplex_count_dict[bc] = 0
    demultiplex_dict["other"] = []  # unassigned barcodes
    demultiplex_count_dict["other"] = 0
    # For each line in the FASTQ file:
    #   Assign to barcode (fastq record into the dictionary)
    #   Cache by barcode; write to the appropriate output files (untrimmed and trimmed)
    #   Use triple brackets in f-strings to leave a single bracket for regex
    pattern = re.compile(
        f"""
        ^                                     #   beginning of string
        ([ACGT]{{{settings.barcode_length}}}) #   group(1) = barcode, any bp (4-16bp) of mixed ACGT
        ([NACGT][ACGT]{{13,14}}               #   group(2) = 16-17 bp of chromosmal sequence
        (?:TA)){settings.transposon_seq}""",  #   flanking transposon sequence, last two must be TA for transposon
        re.VERBOSE,
    )

    # Initiate progress bar
    logger.info("Preparing to demultiplex reads")
    num_fastq_lines = count_lines(reads, fastq=True)
    tqdm_logger = TqdmBarLogger(
        logger, num_fastq_lines, desc="Demultiplexing reads", unit="reads"
    )
    # count of reads
    n_reads = 0
    for read in screed.open(reads):
        tqdm_logger.update()
        m = re.search(pattern, read.sequence)
        try:
            barcode, chrom_seq = m.group(1), m.group(2)
            if barcode in demultiplex_dict:
                read["trim"] = m.span(2)  # trim slice
                demultiplex_dict[barcode].append(read)
            else:
                demultiplex_dict["other"].append(read)
        except:
            demultiplex_dict["other"].append(read)
        # Every 10^6 sequences write and clear the dictionary
        n_reads += 1
        # Buffer dumping reads into barcode file
        if n_reads % 5e6 == 0:
            tqdm_logger.info(f"Demultiplexed {n_reads:,} reads. Writing to file")
            write_reads(demultiplex_dict, samples_dict, settings)
            # Clear the dictionary after writing to file
            for barcode in demultiplex_dict:
                # Record amount of reads in barcode
                demultiplex_count_dict[barcode] += len(demultiplex_dict[barcode])
                demultiplex_dict[barcode] = []
    write_reads(demultiplex_dict, samples_dict, settings)
    # Clear the dictionary after writing to file
    for barcode in demultiplex_dict:
        # Record amount of reads in barcode
        demultiplex_count_dict[barcode] += len(demultiplex_dict[barcode])
        demultiplex_dict[barcode] = []
    # Remove any traces of Progress Bar
    tqdm_logger.p_bar.close()
    # print demultiplex summary
    pyinseq_logger.logger_io.write(f"{n_reads} reads demultiplexed\n")
    for bc in demultiplex_count_dict:
        for sample in samples_dict:
            if samples_dict[sample]["barcode"] == bc:
                samples_dict[sample]["demultiplexed_reads"] = demultiplex_count_dict[bc]
                pyinseq_logger.logger_io.write(
                    f"- {sample} ({bc}): {demultiplex_count_dict[bc]:,} reads"
                )
    logger.info(f"Total reads demultiplexed: {n_reads:,}")

    # Return an update samples_dict
    return samples_dict


def write_reads(demultiplex_dict, samples_dict, settings):
    """Write the fastq data to the correct (demultiplexed) file."""
    # inverting the value: key pairs to barcode: sample, and also adding 'other': '_other'
    barcode_dict = {"other": "_other"}
    for sample in samples_dict:
        barcode_dict[samples_dict[sample]["barcode"]] = sample
    # Write to sample file
    for barcode in demultiplex_dict:
        if demultiplex_dict[barcode]:
            with open(
                settings.raw_path.joinpath(f"{barcode_dict[barcode]}.fastq"), "a"
            ) as fo:
                logger.debug(f"Writing reads to {barcode_dict[barcode]}.fastq")
                # Initialize Progress Bar
                num_lines_to_write = len(demultiplex_dict[barcode])
                tqdm_write_logger = TqdmBarLogger(
                    logger,
                    num_lines_to_write,
                    desc=f"Writing reads to {barcode_dict[barcode]}.fastq",
                    unit="reads",
                )
                for read in demultiplex_dict[barcode]:
                    tqdm_write_logger.update()
                    fo.write(f"@{read.name}\n{read.sequence}\n+\n{read.quality}\n")
                tqdm_write_logger.close()
        # Write trimmed if necessary
        if settings.write_trimmed_reads and barcode != "other":
            with open(
                settings.path.joinpath(f"{barcode_dict[barcode]}_trimmed.fastq"), "a"
            ) as fo:
                logger.debug(
                    f"Writing trimmed reads to {barcode_dict[barcode]}_trimmed.fastq"
                )
                # Same logger, TODO: could maybe wrap loop into iterator
                tqdm_write_logger = TqdmBarLogger(
                    logger,
                    num_lines_to_write,
                    desc=f"Writing trimmed reads to {barcode_dict[barcode]}_trimmed.fastq",
                    unit="reads",
                )
                for read in demultiplex_dict[barcode]:
                    tqdm_write_logger.update()
                    fo.write(
                        f"@{read.name}\n"
                        f"{read.sequence[slice(read.trim[0], read.trim[1])]}\n"
                        f"+\n"
                        f"{read.quality[slice(read.trim[0], read.trim[1])]}\n"
                    )
                tqdm_write_logger.close()
    return


def main():
    """Start here."""


if __name__ == "__main__":
    main()



#!/usr/bin/env/python3

"""

Settings for running baseline pyinseq

"""

import os
import yaml
from pathlib import Path
from pyinseq.logger import pyinseq_logger as logger
from pyinseq.workflows import get_workflow_snakefile_path
from pyinseq.utils import (
    convert_to_filename,
    tab_delimited_samples_to_dict,
    get_config_dict,
)


class Settings:
    """General settings for pyinseq"""

    def __init__(self, config_file):
        # Attributes for all workflows
        self.config_file = config_file
        config_dict = get_config_dict(self.config_file)
        # command options are: ['pyinseq', 'demultiplex', 'genomeprep]
        self.command = config_dict["command"]
        self.experiment = convert_to_filename(config_dict["experiment"])
        self.threads = config_dict["threads"]
        self.path = Path(f"results/{self.experiment}/")
        self.log = self.path.joinpath("log.txt")
        self.summary_log = self.path.joinpath("summary_log.txt")
        self.settings_pickle = self.path.joinpath("settings.pickle")

        # Set snakemake shell command
        self.snakefile = get_workflow_snakefile_path(self.command)
        self.snakemake_cmd = [
            "snakemake",
            "--config",
            f"experiment={self.experiment}",
            "-s",
            str(self.snakefile),
            "--cores",
            str(self.threads - 1),
            "--quiet",
        ]
        # Add additional parameters, if any...
        self.snakemake_cmd.extend(config_dict["snakemake_params"])
        self.snakemake_cmd = " ".join(self.snakemake_cmd)

    def __repr__(self):
        # Print each variable on a separate line
        return "\n".join("%s: %s" % item for item in vars(self).items())

    def clean_up(self):
        """Cleans up settings pickle file"""
        try:
            os.remove(self.settings_pickle)
        except OSError:
            pass
        # combine logging text here
        with open(self.summary_log, "r") as f:
            summary_log_text = "".join(f.readlines())
        self.summary_log.unlink()
        with open(self.log, "a") as f:
            f.write("-" * 50 + "\n\n")
            f.write(summary_log_text)

    def dump_sample_dict_to_yml(self, updated_dict=None):
        """Updates sample_info.yml file with attributes for sample"""
        if not updated_dict:
            # Dump yml file for info
            with open(self.samples_info_yml, "w") as f:
                yaml.dump(self.samples_dict, f)
            return
        # Load current state of sample_dict
        with open(self.samples_info_yml, "r+") as f:
            current_dict = yaml.load(f, Loader=yaml.FullLoader)
        # Updated the old sample with new information
        with open(self.samples_info_yml, "w") as fo:
            dump_dict = current_dict.copy()
            for sample, sub_d in updated_dict.items():
                for k in sub_d:
                    if k not in current_dict[sample].keys():
                        dump_dict[sample][k] = updated_dict[sample][k]
            yaml.dump(dump_dict, fo)
        return


class PyinseqSettings(Settings):
    def __init__(self, config_file):
        super().__init__(config_file)
        config_dict = get_config_dict(self.config_file)
        self.reads = config_dict["input"]
        self.samples = config_dict["samples"]
        self.samples_txt = self.path.joinpath("samples.txt")
        self.samples_dict = tab_delimited_samples_to_dict(self.samples)
        self.samples_info_yml = self.path.joinpath("samples_info.yml")
        self.raw_path = self.path.joinpath("raw_data")
        self.reference_genome = config_dict["genome"]
        self.genome_path = self.path.joinpath("genome_lookup")
        # organism reference files called 'genome.fna' etc
        self.organism = "genome"
        self.gff3 = config_dict["gff3"]
        self.summary_table = self.path.joinpath("summary_gene_table.txt")
        # Pyinseq optional args
        self.disruption = set_disruption(config_dict["disruption"])
        self.barcode_length = set_barcode_length(config_dict["barcode_length"])
        self.transposon_seq = set_transposon_seq(config_dict["transposon_seq"])
        # counts at one transposon site for it to qualify
        self.min_counts = set_min_count(config_dict["min_count"])
        # max ratio of left/right sites for it to qualify
        self.max_ratio = set_max_ratio(config_dict["max_ratio"])
        # Actions to carry out
        self.map_to_genome = True
        self.process_reads = True
        self.process_sample_list = True
        self.parse_genebank = True
        self.generate_bowtie_index = True
        self.write_trimmed_reads = True


class DemultiplexSettings(Settings):
    def __init__(self, config_file):
        super().__init__(config_file)
        config_dict = get_config_dict(self.config_file)
        self.reads = config_dict["input"]
        self.samples = config_dict["samples"]
        self.samples_txt = self.path.joinpath("samples.txt")
        self.samples_dict = tab_delimited_samples_to_dict(self.samples)
        self.samples_info_yml = self.path.joinpath("samples_info.yml")
        self.raw_path = self.path.joinpath("raw_data")
        self.barcode_length = set_barcode_length(config_dict["barcode_length"])
        self.transposon_seq = set_transposon_seq(config_dict["transposon_seq"])
        # Actions to carry out
        self.parse_genebank = False
        self.process_reads = True
        self.process_sample_list = True
        self.write_trimmed_reads = not config_dict["notrim"]


class GenomeprepSettings(Settings):
    def __init__(self, config_file):
        super().__init__(config_file)
        config_dict = get_config_dict(self.config_file)
        self.reference_genome = config_dict["genome"]
        self.genome_path = self.path.joinpath("genome_lookup")
        self.gff3 = config_dict["gff3"]
        # organism reference files called 'genome.fna' etc
        self.organism = "genome"
        # Actions to carry out
        self.parse_genebank = True
        self.generate_bowtie_index = True
        self.process_reads = False


# HOLDS CLASS CONSTRUCTORS
SETTINGS_CONSTRUCTORS = {
    "pyinseq": PyinseqSettings,
    "demultiplex": DemultiplexSettings,
    "genomeprep": GenomeprepSettings,
}


def set_min_count(min_count=3):
    """Check min_count is positive, otherwise set default."""
    if min_count < 0:
        logger.error(
            f"Min_count value provided ({min_count}) is not positive; proceeding with default value of 3"
        )
        return 3
    return min_count


def set_max_ratio(max_ratio=10):
    """Check that max_ratio is positive, otherwise set default."""
    if max_ratio < 0:
        logger.error(
            f"Max_ratio value provided ({max_ratio}) is not positive; proceeding with default value of 10"
        )
        return 10
    return max_ratio


def set_disruption(d=0.9):
    """Check that gene disruption is 0.0 to 1.0; otherwise set to 0.9."""
    if d < 0.0 or d > 1.0:
        logger.error(
            f"Disruption value provided ({d}) is not in range 0.0 to 0.9; proceeding with default value of 0.9"
        )
        return 0.9
    return d


def set_barcode_length(barcode_length=4):
    """Check that min_count and max_ratio are positive, otherwise set default."""
    if barcode_length > 16:
        logger.error(
            f"Barcode length {barcode_length} is larger than 16bp which is unsupported by pyinseq. "
            f"Setting length to 4bp default"
        )
        return 4
    if barcode_length < 4:
        logger.error(
            f"Barcode length {barcode_length} is smaller than 4bp which is unsupported by pyinseq. "
            f"Setting length to 4bp default"
        )
        return 4
    return barcode_length


def set_transposon_seq(transposon_seq="ACAGGTTG"):
    """Check that min_count and max_ratio are positive, otherwise set default."""
    # TODO: Check that they are DNA nucleotides ATGC
    if not transposon_seq.isalpha():
        logger.error(
            f"Transposon sequence provided contains non-letter characters. "
            f"Setting default to ACAGGTTG"
        )
        return "ACAGGTTG"
    return transposon_seq



#!/usr/bin/env python3

"""

Custom logger for pyinseq

"""

import io
import sys
import tqdm
import logging


class CustomLoggers:

    FORMATTER = logging.Formatter(
        fmt="{asctime} - {levelname} - {module} - {message}",
        datefmt="%Y-%m-%d %H:%M",
        style="{",
    )

    def __init__(self):
        # Set io's
        self.logger_io = io.StringIO()
        self.snake_io = io.StringIO()

    def add_logfile_paths(self, settings):
        self.log = settings.log
        self.summary_log = settings.summary_log

    def summarize_step(self):
        """Dumps io buffer into a log summary file"""
        with open(self.summary_log, "a+") as f:
            snake_message = self.snake_io.getvalue()
            pyinseq_message = self.logger_io.getvalue()
            f.write(snake_message)
            f.write("(PYINSEQ INFO)\n" + pyinseq_message)
        # Flush io's
        self.snake_io.flush()
        self.logger_io.flush()

    def setup_logger(self, formatter=None):
        """Sets up logger with sys.stdout handler"""
        self.logger = logging.getLogger("pyinseq")
        self.logger.setLevel(logging.INFO)
        stream_handler = logging.StreamHandler(stream=sys.stdout)
        if formatter:
            stream_handler.setFormatter(formatter)
        else:
            stream_handler.setFormatter(CustomLoggers.FORMATTER)
        stream_handler.setLevel(logging.INFO)
        self.logger.addHandler(stream_handler)
        return


## ONLY LOGGER IN PYINSEQ ##
pyinseq_logger = CustomLoggers()


class TqdmBarLogger:
    """

    Custom progress bar that works with Logger

    """

    def __init__(
        self, logger: logging.Logger, num_iterations: int, desc: str, unit: str
    ):
        self.logger = logger
        self.p_bar = tqdm.tqdm(
            total=num_iterations,
            desc=desc,
            unit=unit,
            leave=False,
            position=0,
            colour="green",
        )

    def info(self, msg):
        """Removes logger from display and logs message"""
        self.p_bar.close()
        self.logger.info(msg)
        # Set disable to False, brings bar back
        self.p_bar.disable = False

    def update(self):
        self.p_bar.update()

    def close(self):
        self.p_bar = None


def add_fileHandler(logger, logfile, formatter=None):
    """Adds a filehandler to logger"""
    fh = logging.FileHandler(logfile, "a")
    if formatter:
        fh.setFormatter(formatter)
    else:
        fh.setFormatter(CustomLoggers.FORMATTER)
    logger.addHandler(fh)
    return


def add_streamHandler(logger, stream, formatter=False):
    """Adds a stream handler to logger"""
    stream_handler = logging.StreamHandler(stream=stream)
    if formatter:
        stream_handler.setFormatter(CustomLoggers.FORMATTER)
    stream_handler.setLevel(logging.INFO)
    logger.addHandler(stream_handler)
    return



#!/usr/bin/env python3

"""

Tests for running pyinseq pipelines

"""

from .test_utils import runscript, datadir, compare_directories


def test_pyinseq_script_no_args(datadir, tmpdir):
    args = []
    status, out, err = runscript("pyinseq", args, directory=str(tmpdir))
    assert out.split("\n")[1][0:5] == "usage"


def test_pyinseq_script(datadir, tmpdir):
    input_fn = datadir("input/example01.fastq")
    sample_fn = datadir("input/example01.txt")
    gb_fn = datadir("input/ES114v2.gb")
    output_name = "example_pyinseq"
    expected_output = datadir("output_pyinseq")
    output_dir = tmpdir.join("results/example_pyinseq")

    args = [
        "-i",
        input_fn,
        "-s",
        sample_fn,
        "-g",
        gb_fn,
        "-e",
        output_name,
        "-d",
        "0.9",
        "--snakemake_params",
        "--use-conda",
    ]
    status, out, err = runscript("pyinseq", args, directory=str(tmpdir))
    assert not status

    # Compare directory outputs
    compare_directories(
        expected_output,
        output_dir,
        ignore=[
            "E001_01_bowtie.txt",
            "E001_02_bowtie.txt",
            ".DS_Store",
            "log.txt",
            "summary_log.txt",
        ],
    )


def test_pyinseq_script_unique_transp_barcode_length(datadir, tmpdir):
    input_fn = datadir("input/example03.fastq")
    sample_fn = datadir("input/example03.txt")
    gb_fn = datadir("input/ES114v2.gb")
    output_name = "example_pyinseq_unique_transposon_barcode"
    expected_output = datadir("output_pyinseq_unique_transposon_barcode_length")
    output_dir = tmpdir.join("results/example_pyinseq_unique_transposon_barcode")

    args = [
        "-i",
        input_fn,
        "-s",
        sample_fn,
        "-g",
        gb_fn,
        "-e",
        output_name,
        "-d",
        "0.9",
        "--barcode_length",
        "9",
        "--transposon_seq",
        "TCGCACGG",
        "--snakemake_params",
        "--use-conda",
    ]
    status, out, err = runscript("pyinseq", args, directory=str(tmpdir))
    assert not status

    # Compare directory outputs
    compare_directories(
        expected_output,
        output_dir,
        ignore=[
            "E001_01_bowtie.txt",
            "E001_02_bowtie.txt",
            ".DS_Store",
            "log.txt",
            "summary_log.txt",
        ],
    )


def test_pyinseq_demultiplex_script(datadir, tmpdir):
    input_fn = datadir("input/example01.fastq")
    sample_fn = datadir("input/example01.txt")
    output_name = "test_demultiplex"
    expected_output = datadir("output_demultiplex")
    output_dir = tmpdir.join(f"results/{output_name}")

    args = ["demultiplex", "-i", input_fn, "-s", sample_fn, "-e", output_name]
    status, out, err = runscript("pyinseq", args, directory=str(tmpdir))
    assert not status

    # Compare directory outputs
    compare_directories(
        expected_output, output_dir, ignore=[".DS_Store", "log.txt", "summary_log.txt"]
    )


def test_pyinseq_demultiplex_notrim_script(datadir, tmpdir):
    input_fn = datadir("input/example01.fastq")
    sample_fn = datadir("input/example01.txt")
    output_name = "test_demultiplex_notrim"
    expected_output = datadir("output_demultiplex_notrim")
    output_dir = tmpdir.join(f"results/{output_name}")

    args = [
        "demultiplex",
        "-i",
        input_fn,
        "-s",
        sample_fn,
        "-e",
        output_name,
        "--notrim",
    ]
    status, out, err = runscript("pyinseq", args, directory=str(tmpdir))
    assert not status

    # Compare directory outputs
    compare_directories(
        expected_output, output_dir, ignore=[".DS_Store", "log.txt", "summary_log.txt"]
    )


def test_pyinseq_genomeprep_script(datadir, tmpdir):
    output_name = "test_genomeprep"
    expected_output = datadir("output_genomeprep")
    output_dir = tmpdir.join(f"results/{output_name}")
    gb_fn = datadir("input/ES114v2.gb")

    args = [
        "genomeprep",
        "-e",
        output_name,
        "-g",
        gb_fn,
        "--snakemake_params",
        "--use-conda",
    ]
    status, out, err = runscript("pyinseq", args, directory=str(tmpdir))
    assert not status

    # Compare directory outputs
    compare_directories(expected_output, output_dir, ignore=[".DS_Store", "log.txt"])


def test_pyinseq_genomeprep_gff_script(datadir, tmpdir):
    output_name = "test_genomeprep_gff"
    expected_output = datadir("output_genomeprep_gff")
    output_dir = tmpdir.join(f"results/{output_name}")
    gb_fn = datadir("input/ES114v2.gb")

    args = [
        "genomeprep",
        "-e",
        output_name,
        "-g",
        gb_fn,
        "--gff",
        "--snakemake_params",
        "--use-conda",
    ]
    status, out, err = runscript("pyinseq", args, directory=str(tmpdir))
    assert not status

    # Compare directory outputs
    compare_directories(expected_output, output_dir, ignore=[".DS_Store", "log.txt"])



#!/usr/bin/env python3

"""

Tests for individual functions in pyinseq module

"""

import os
import sys
import filecmp
from pathlib import Path
from collections import OrderedDict

# Setup logger before importing modules
from pyinseq.logger import pyinseq_logger

pyinseq_logger.setup_logger()
# Module imports
from pyinseq import utils
from pyinseq.settings import Settings
from pyinseq.parsers import get_args

# Test imports
from .test_utils import cd, runscript, datadir, load_settings, get_dump


def test_filename_replace_spaces():
    assert utils.convert_to_filename("file name") == "file_name"


def test_filename_leading_whitespace():
    assert utils.convert_to_filename("  filename") == "filename"


def test_filename_trailing_whitespace():
    assert utils.convert_to_filename("filename  ") == "filename"


def test_filename_funny_characters():
    assert utils.convert_to_filename("filename*&*&*$$") == "filename"
    assert utils.convert_to_filename("file*&*&*$$name") == "filename"


def test_pyinseq_Settings(datadir, load_settings):
    setting = load_settings("pyinseq")

    assert setting.path == Path("results/test_pyinseq/")
    assert setting.genome_path == Path("results/test_pyinseq/genome_lookup/")
    assert setting.raw_path == Path("results/test_pyinseq/raw_data/")
    assert setting.samples_txt == Path("results/test_pyinseq/samples.txt")
    assert setting.log == Path("results/test_pyinseq/log.txt")
    assert setting.summary_log == Path("results/test_pyinseq/summary_log.txt")
    assert setting.barcode_length == 4


def test_tab_delimited_samples_to_dict_trailing_newline():
    s = "pyinseq/tests/data/additional/sample01_01.txt"
    assert utils.tab_delimited_samples_to_dict(s) == OrderedDict(
        [("sample_1", {"barcode": "AAAA"}), ("sample_2", {"barcode": "TTTT"})]
    )


def test_tab_delimited_samples_to_dict_no_trailing_newline():
    s = "pyinseq/tests/data/additional/sample01_02.txt"
    assert utils.tab_delimited_samples_to_dict(s) == OrderedDict(
        [("sample_1", {"barcode": "AAAA"}), ("sample_2", {"barcode": "TTTT"})]
    )


def test_create_experiment_directories(datadir, tmpdir, load_settings):
    """Test for function that creates the directory structure of pyinseq"""
    settings = load_settings("pyinseq")
    expected_output = datadir("output_pyinseq")
    # In the context of pytest directory
    with cd(str(tmpdir)):
        # Manually modify settings object
        from pathlib import Path

        settings.path = Path(tmpdir + "results/test_pyinseq")
        utils.create_experiment_directories(settings)
        # Check /results/experiment exist
        assert os.path.exists(settings.path)
        assert filecmp.dircmp(expected_output, settings.path).common_dirs == [
            "genome_lookup",
            "raw_data",
        ]


def test_write_config_file(datadir, tmpdir):
    """Test for writing configuration file"""
    input_fn = "../data/input/example01.fastq"
    sample_fn = "../data/input/example01.txt"
    gb_fn = "../data/input/ES114v2.gb"
    output_name = "test_pyinseq"
    expected_config = datadir("input/test-pyinseq-config.yml")
    output_config = tmpdir.join("test_pyinseq-config.yml")

    # Modify sys.argv
    sys.argv = [
        "pyinseq",
        "-i",
        input_fn,
        "-s",
        sample_fn,
        "-g",
        gb_fn,
        "-e",
        output_name,
        "-d",
        "0.9",
        "--threads",
        "2",
        "--snakemake_params",
        "--use-conda",
    ]
    # Get the Namespace object from argparse
    parser, args = get_args()
    with cd(str(tmpdir)):
        utils.write_config_file(args)
        # Check files
        assert filecmp.cmp(expected_config, output_config)



#!/usr/bin/env python3

"""

GenBank conversion utilities for the pyinseq pipeline.

Files are written to genome_lookup/ directory for the EXPERIMENT:
    EXPERIMENT/genome_lookup/genome.fna

# gbk2fna()
Convert GenBank sequence to fasta sequence
Multilocus GenBank converts to one multifasta GenBank file
Locus headers are the fasta headers
Maintains original newlines (typically leaving up to 60 nucleotides per line)

# gbk2table()


"""

import re
import csv

# Module imports
from pyinseq.logger import pyinseq_logger

logger = pyinseq_logger.logger


def build_fna_and_table_files(gbk_file, settings):
    """Convert GenBank file to a fasta nucleotide (.fna) and feature table (.ftt) files."""
    logger.info(
        f"Converting GenBank {gbk_file} to fasta nucleotide (.fna) and feature table (.ftt)."
    )
    fasta = gbk2fna(gbk_file, settings.organism, settings.genome_path)
    gbk2table(gbk_file, fasta, settings.organism, settings.genome_path, settings.gff3)


def write_to_file(outfile, row_data):
    """Writes table into outfile"""
    with open(outfile, "w") as fo:
        writer = csv.writer(fo, delimiter="\t", lineterminator="\n")
        for row in row_data:
            writer.writerow(row)
    return


def gbk2fna(infile, organism, output_directory=""):
    """Convert genbank format to fna format."""
    with open(infile, "r") as fi:
        fna_rows = []
        n_nt = {}
        dna_seq = False  # in the DNA sequence of the file
        for i, line in enumerate(fi):

            # Don't parse blank lines
            if line.strip():
                parts = line.split()

                # Locus (replicon) as header
                if parts[0] == "LOCUS":
                    locus = parts[1]
                    n_nt[locus] = 0
                    fna_rows.append((f">{locus}",))

                # DNA Sequence
                if parts[0] == "//":
                    dna_seq = False
                if dna_seq:
                    sequence = "".join(n for n in line.strip() if n.isalpha())
                    fna_rows.append((f"{sequence}",))
                    n_nt[locus] += len(sequence.strip())
                if parts[0] == "ORIGIN":
                    dna_seq = True

        outfile = output_directory.joinpath(f"{organism}.fna")
        write_to_file(outfile, fna_rows)

    logger.info(f"Nucleotides stored in {outfile}")
    return {"fasta": fna_rows, "nucleotides": n_nt}


def gbk2table(infile, fasta, organism, output_directory="", gff3=False):
    """Convert GenBank to feature table
    Format similar to .ptt and .rnt files except:
    - full tabular (locus as a field)
    - start and end positions as separate fields
    - includes the following features:
        CDS
        rRNA
        tRNA
        misc_RNA
    - Multilocus GenBank converts to multi-.ftt file
    'Unlike .ptt files that show the number of amino acids as 'length'
    Optional convert to GFF3 format also..
    """
    with open(infile, "r") as fi:
        ftt_rows = [
            (
                "Locus",
                "Location_Start",
                "Location_End",
                "Strand",
                "Length",
                "PID",
                "Gene",
                "Synonym",
                "Code",
                "COG",
                "Product",
            )
        ]

        # Collect/write fasta sequence and collect contig lengths
        fna_rows = fasta["fasta"]
        fna_nucleotides = fasta["nucleotides"]

        if gff3:
            logger.info("Writing gff3 version of genome files")
            gff_rows = [("##gff-version 3",)]
            for contig in fna_nucleotides:
                gff_rows.append(
                    (f"##sequence-region {contig} 1 {fna_nucleotides[contig]}",)
                )

        # Initialize variables
        features = False  # in the FEATURES section of the GenBank file
        new_feature = False  # collecting data for a new feature
        parse_types = ["CDS", "tRNA", "rRNA", "misc_RNA"]
        feature_type = "CDS"
        strand = "+"
        length = 0
        protein_id = "-"
        gene = "-"
        locus_tag = "-"
        code = "-"
        cog = "-"
        product = "-"

        product_append = False  # append the current line to product
        """peptide_append = False         
           peptide_sequence = '-'  # append the current line to peptide
        """

        for i, line in enumerate(fi):

            # Don't parse blank lines
            if line.strip():
                parts = line.split()

                # PRINT HEADER FOR THE LOCUS
                # 2 LINES:
                # LOCUS <tab> locus name
                # Location <tab> Strand etc...
                if parts[0] == "LOCUS":
                    locus = parts[1]

                if features:
                    # Print line before go on to next feature
                    # (gene, COG, protein id not required)
                    # Reset flags/defaults
                    if line[5:21].rstrip():
                        if new_feature:
                            # if locus_tag:
                            if not product_append:
                                ftt_rows.append(
                                    (
                                        locus,
                                        first,
                                        last,
                                        strand,
                                        str(length),
                                        protein_id,
                                        gene,
                                        locus_tag,
                                        code,
                                        cog,
                                        product,
                                    )
                                )
                                if gff3:
                                    gff_id = (
                                        f"ID:{locus_tag}"
                                        if gene == "-"
                                        else f"ID:{locus_tag};Name:{gene}"
                                    )
                                    gff_rows.append(
                                        (
                                            locus,
                                            ".",
                                            feature_type,  # need to extract the type of feature
                                            first,
                                            last,
                                            ".",
                                            strand,
                                            ".",
                                            gff_id,
                                        )
                                    )

                                new_feature = False

                    if line[5:21].rstrip() in parse_types:
                        new_feature = True  # Feature that should be written
                        feature_type = line[5:21].rstrip()
                        protein_id = "-"
                        gene = "-"
                        locus_tag = "-"
                        code = "-"
                        cog = "-"
                        product = "-"

                        # NOTES ABOUT FEATURES
                        # 1. At ends of contigs greater than/less than signs
                        #    (> / <) are removed.
                        # 2. Complicated features use only the outer bounds
                        #    join(481257..481331,481333..482355) uses 481257..482355
                        location = re.search(r"(\d+)\.+.*\.(\d+)", parts[1])
                        if not location:
                            print(location)
                            print(line)
                            continue

                        first = location.group(1)
                        last = location.group(2)
                        try:
                            length = int(last) - int(first) + 1
                        except AttributeError:
                            error_complex_feature = (
                                f"PyINSeq Error: Complex feature coordinates at or near {locus_tag} "
                                "in GenBank file. Additional attention required."
                            )
                            print(error_complex_feature)
                            exit(0)
                        strand = "-" if parts[1].startswith("complement") else "+"

                    if "/protein_id=" in parts[0]:
                        protein_id = parts[0][13:-1]

                    if "/gene=" in parts[0]:
                        gene = parts[0][7:-1]

                    if "/locus_tag=" in parts[0]:
                        locus_tag = parts[0][12:-1]

                    # Multi-line product description
                    if product_append:
                        product = product + " " + line.strip()
                        if product.count('"') != 2:
                            product_append = True
                        # TODO(Error handling if not exactly 2 parentheses)
                        if product.count('"') == 2:
                            product = product.strip('"')
                            product_append = False

                    if "/product=" in parts[0]:
                        product = line.strip()[9:]
                        if product.count('"') != 2:
                            product_append = True
                        if product.count('"') == 2:
                            product = product.strip('"')

                    """if peptide_append:
                        peptide_sequence += line.strip()
                    if peptide_sequence.count('"') == 2:
                            peptide_sequence = peptide_sequence.strip('"')
                            peptide_append = False
                    # TODO: Add this as an option for including amino acid sequences, leave as comment
                    if "/translation=" in parts[0]:
                        peptide_sequence = parts[0].replace('/translation=', '').strip()
                        if parts[0].count('"') != 2:
                            peptide_append = True"""

                if parts[0] == "ORIGIN":
                    features = False  # Not in FEATURES any more
                if parts[0] == "FEATURES":
                    features = True

    outfile = output_directory.joinpath(f"{organism}.ftt")
    write_to_file(outfile, ftt_rows)
    if gff3:
        gff_rows.append(("##FASTA",))
        gff_rows = gff_rows + fna_rows
        outfile = output_directory.joinpath(f"{organism}.gff")
        write_to_file(outfile, gff_rows)
    logger.info(f"Features table stored in {outfile}")
    return


if __name__ == "__main__":
    pass



#!/usr/bin/env python
"""
Analyzes resulting output
"""
import numpy as np
import pandas as pd
import seaborn as sn
import matplotlib.pyplot as plt


def read_sites_file(sample: str, settings: "runner.Settings") -> pd.DataFrame:
    return pd.read_csv(f"{settings.path}{sample}_sites.txt", sep="\t")


def read_summary_table(settings: "runner.Settings") -> pd.DataFrame:
    return pd.read_csv(settings.summary_table, sep="\t")


def t_fifty(sample: str, settings: "runner.Settings") -> pd.DataFrame:
    """T50: The minimum number of transposon insertion sites in the sample that
    account for at least 50% of the samples's reads.
    """
    # calculate sum of counts in df
    df = read_sites_file(sample, settings)
    return np.sum(df.total_counts.sort_values().cumsum() > (df.total_counts.sum() / 2))


def spearman_correlation(samples: dict, settings: "runner.Settings") -> pd.DataFrame:
    """Calculates spearman correlation among samples and plots heatmap"""
    df = read_summary_table(settings)
    corr_df = df[list(samples.keys())].corr("spearman")
    corr_df.to_csv(f"{settings.analysis_path}correlation.csv", sep="\t")
    plot_heatmap(corr_df)
    return


def plot_insertions(samples: dict, settings: "runner.Settings") -> pd.DataFrame:
    """Plot insertions across each contig/chromosome."""
    for sample in samples.keys():

        sites_df = read_sites_file(sample, settings)

        for i, group in sites_df.groupby("contig"):
            fig, ax = plt.subplots()
            group.plot(
                kind="scatter",
                x="nucleotide",
                y="cpm",
                color="blue",
                alpha=0.5,
                title=str(i),
                ax=ax,
            )
            fig.savefig(
                f"{settings.figures_path}insertions_scatter_{sample}_{str(i)}.pdf",
                format="pdf",
            )
            return


def plot_heatmap(df):
    pass


def plot_MA(sample, settings):
    pass


def plot_pairwise(samples, settings):
    pass


def df_raw_counts(samples: dict, settings: "runner.Settings") -> pd.DataFrame:
    """Returns total raw counts from samples in a DataFrame."""
    # Results DataFrame with desired columns
    result_df = pd.DataFrame(
        columns=["contig", "nucleotide", "left_counts", "right_counts", "total_counts"]
    )
    col = result_df.columns[2:5]

    # Loop over samples and add cell values to result_df
    for sample in samples.keys():
        df = pd.read_csv(f"{settings.path}{sample}_sites.txt", sep="\t").drop(
            "cpm", axis=1
        )
        for i, row in df.iterrows():
            if row["nucleotide"] in result_df.nucleotide.tolist():
                for c in col:
                    previous = result_df.loc[
                        row["nucleotide"] == result_df["nucleotide"], c
                    ].item()
                    result_df.loc[row["nucleotide"] == result_df["nucleotide"], c] = (
                        previous + row[c]
                    )
            else:
                result_df = result_df.append(row)
    result_df.to_csv(
        f"{settings.analysis_path}{settings.experiment}_raw_counts.csv", sep="\t"
    )
    return


def df_cpm_counts(samples: dict, settings: "runner.Settings") -> pd.DataFrame:
    """Returns total cpm counts from samples in a DataFrame."""
    # Results DataFrame with desired columns
    result_df = pd.DataFrame(columns=["contig", "nucleotide", "cpm"])

    # Loop over samples and add cell values to result_df
    for sample in samples.keys():
        df = pd.read_csv(
            f"{settings.path}{sample}_sites.txt", sep="\t", usecols=[0, 1, 5]
        )
        for i, row in df.iterrows():
            if row["nucleotide"] in result_df.nucleotide.tolist():
                previous = result_df.loc[
                    row["nucleotide"] == result_df["nucleotide"], "cpm"
                ].item()
                result_df.loc[row["nucleotide"] == result_df["nucleotide"], "cpm"] = (
                    previous + row["cpm"]
                )
            else:
                result_df = result_df.append(row)
    result_df.to_csv(
        f"{settings.analysis_path}{settings.experiment}_cpm_counts.csv", sep="\t"
    )
    return


def df_gene_counts(samples: dict, settings: "runner.Settings") -> pd.DataFrame:
    """Return total count of hits per gene."""
    # Results DataFrame with desired columns
    results_df = pd.DataFrame(
        columns=[
            "contig",
            "left_counts",
            "right_counts",
            "total_counts",
            "cpm",
            "locus_tag",
        ]
    )
    col = results_df.columns[1:5]

    # Loop over samples and add cell values to result_df
    for sample in samples.keys():
        # TODO: caclulate average of three_primeness of gene across samples
        # Add gene index to results_df
        df = pd.read_csv(f"{settings.path}{sample}_genes.txt", sep="\t").drop(
            ["nucleotide", "three_primeness"], axis=1
        )
        for i, row in df.iterrows():
            if row["locus_tag"] in results_df.locus_tag.tolist():
                for c in col:
                    previous = results_df.loc[
                        results_df["locus_tag"] == row["locus_tag"], c
                    ].item()
                    results_df.loc[results_df["locus_tag"] == row["locus_tag"], c] = (
                        previous + row[c]
                    )
            else:
                results_df = results_df.append(row)
    results_df.to_csv(
        f"{settings.analysis_path}{settings.experiment}_gene_counts.csv", sep="\t"
    )

    return


def main():
    """Start here."""
    samples = f"results/{'example02'}/samples.yml"
    with open(samples, "r") as f:
        organize_samples(f)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import os
import sys
from pathlib import Path

try:
    from setuptools import *
except ImportError:
    from distribute_setup import use_setuptools

    use_setuptools()
finally:
    from setuptools import *

from glob import glob

if sys.version_info < (3, 6):
    print >>sys.stderr, "ERROR: pyinseq requires python 3.6 or greater"
    sys.exit()

__version__ = open(os.path.join("pyinseq", "VERSION")).read().strip()

SCRIPTS = glob("scripts/*")

with open("README.md", "r") as fh:
    long_description = fh.read()


# Collect paths to Snakefile's and ENVS
snake_data = [
            "pyinseq/workflows/PyinseqWorkflow/Snakefile",
            "pyinseq/workflows/DemultiplexWorkflow/Snakefile",
            "pyinseq/workflows/GenomeprepWorkflow/Snakefile",
            "pyinseq/envs/bowtie.yaml",
]

def main():
    setup(
        name="pyinseq",
        version=__version__,
        description="Analysis of transposon insertion sequencing (INSeq) data in Python",
        long_description=long_description,
        long_description_content_type="text/markdown",
        url="https://github.com/mjmlab/pyinseq",
        author="Mark J. Mandel,Emanuel Burgos,Camille Scott,Loren Velasquez,Benjamin K. Johnson",
        author_email="mandel01@gmail.com",
        license="BSD",
        packages=find_packages(),
        include_package_data=True,
        package_dir={'pyinseq': 'pyinseq'},
        package_data={'pyinseq': snake_data},
        scripts=SCRIPTS,
        setup_requires=["pytest-runner"],
        tests_require=["pytest"],
        python_requires='!=3.8.*,>=3.6.*',
        classifiers=[
            "Development Status :: 4 - Beta",
            "Intended Audience :: Science/Research",
            "License :: OSI Approved :: BSD License",
            "Operating System :: MacOS :: MacOS X",
            "Operating System :: POSIX",
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: 3.7",
            "Topic :: Scientific/Engineering :: Bio-Informatics",
        ],
        zip_safe=False,
    )


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

"""

Functions for using bowtie

"""

import re
import logging
import subprocess

# Module imports
from pyinseq.logger import pyinseq_logger, add_fileHandler

# Custom snake message for jobs
SNAKE_MESSAGE = """
(SNAKEMAKE INFO)
Job counts:
\tcount	jobs
\t1\tbowtie_mapping
\t1
"""

formatter = logging.Formatter(
    fmt="{asctime} - {levelname} - map_reads - {message}",
    datefmt="%Y-%m-%d %H:%M",
    style="{",
)


def bowtie_build(organism):
    """Build a bowtie index given a fasta nucleotide file."""
    fna = organism + ".fna"
    subprocess.check_call(["bowtie-build", "-q", fna, organism])


def bowtie_map(organism, reads, bowtie_output, threads):
    """Map fastq reads to a bowtie index."""
    fna = organism + ".fna"
    # String version of the shell command
    bash_command = (
        f"bowtie -m 1 --best --strata -a --fullref -n 1 -l 17 "
        f"{organism} -q {reads} {bowtie_output} -p {threads}"
    )
    # Convert bash command to run properly - no spaces; instead list of entries
    # that will appear in the shell as space-separated
    # Consider shlex.split() instead of split() -- any benefit here?
    # subprocess.check_call(bashCommand.split(' '))
    logger.info(f"Mapping reads in {reads} to genome using bowtie")
    proc = subprocess.Popen(
        bash_command.split(" "), stdout=subprocess.PIPE, stderr=subprocess.STDOUT
    )
    bowtie_msg_out = proc.communicate()[0]
    # decode from bytes to unicode
    bowtie_msg_out = bowtie_msg_out.decode("utf8")
    return bowtie_msg_out


def parse_bowtie(bowtie_message):
    """Parse bowtie results into a dictionary."""
    bowtie_msg_dict = {}
    for line in bowtie_message.split("\n"):
        # extract counts from bowtie printing
        m = re.search(r"^(\#.+:) (\d+)", line)
        try:
            message, count = m.group(1), int(m.group(2))
            bowtie_msg_dict[message] = count
        except:
            pass
    return bowtie_msg_dict


def summarize_mapping(sample, bowtie_msg_dict):
    """Summarizes bowtie mapping step into pyinseq_logger io"""
    pyinseq_logger.logger_io.write(
        f"- Mapped reads from {sample} to genome\nBOWTIE OUTPUT - {sample}:\n"
    )
    for key, value in bowtie_msg_dict.items():
        pyinseq_logger.logger_io.write(f"{key} {value}")
    return


# ===== Start here ===== #
if __name__ == "__main__":
    # Modified to work with snakemake
    organism = snakemake.params[0]
    reads = snakemake.input[0]
    output = snakemake.output[0]
    threads = snakemake.threads
    sample = snakemake.wildcards[0]

    # Setup logger
    pyinseq_logger.setup_logger(formatter=formatter)
    logger = pyinseq_logger.logger
    pyinseq_logger.summary_log = snakemake.params.summary_log
    add_fileHandler(logger, snakemake.params.log, formatter=formatter)

    # Print and write custom snake message
    print(SNAKE_MESSAGE)
    pyinseq_logger.snake_io.write(SNAKE_MESSAGE)

    # Begin bowtie
    bowtie_msg = bowtie_map(organism, reads, output, threads)
    # Print bowtie results for sample
    logger.info(f"Bowtie results for {sample}:")
    with open(snakemake.params.log, "a+") as f:
        print(bowtie_msg)
        print(bowtie_msg, file=f)

    bowtie_msg_dict = parse_bowtie(bowtie_msg)
    summarize_mapping(sample, bowtie_msg_dict)
    # Writes to log files
    pyinseq_logger.summarize_step()



#!/usr/bin/python

"""

Ad-hoc functions for snakemake workflows

"""

from pathlib import Path


def get_workflow_snakefile_path(command):
    """Returns absolute path for Snakefile"""
    workflows = {
        "pyinseq": Path(__file__).parent.joinpath(f"PyinseqWorkflow/Snakefile"),
        "demultiplex": Path(__file__).parent.joinpath(f"DemultiplexWorkflow/Snakefile"),
        "genomeprep": Path(__file__).parent.joinpath(f"GenomeprepWorkflow/Snakefile"),
    }
    return workflows[command]
